package fr.bhprod.sga.path;

import java.io.BufferedReader;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStreamReader;
import java.util.Date;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Vector;
import java.util.Map.Entry;
import java.util.logging.Level;

import org.apache.log4j.Logger;

import com.jme.app.SimpleGame;
import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.scene.Line;
import com.jme.scene.shape.Quad;
import com.jme.scene.shape.Sphere;
import com.jme.util.LoggingSystem;

/**
 * http://www.supinfo-projects.com/fr/2006/pathfinding%5Ffr%5F2006/introduction/
 * @author jbeaufils
 * @version 1.0
 */

public class TestAStar extends SimpleGame {

	private short[][] walkable = null;
	private Hashtable<String, PathTile> listeOuverte = null;
	private Hashtable<String, PathTile> listeFermee = null;
	private short sizeX = 13;
	private short sizeY = 13;
	private Vector2f start = new Vector2f(0,8);
	private Vector2f dest = new Vector2f(8,10);
	private PathTile tileDest = null;
	public static final int STRAIGHT_COST = 10;
	public static final int DIAG_COST = 14;
	public static final Logger LOGGER = Logger.getLogger(TestAStar.class);
	
	public static void main(String[] args) {
		TestAStar app = new TestAStar();
		app.setDialogBehaviour(FIRSTRUN_OR_NOCONFIGFILE_SHOW_PROPS_DIALOG);
		app.start();
	}

	@Override
	protected void simpleInitGame() {
		LoggingSystem.getLogger().setLevel(Level.OFF);

		display.setTitle("SGA");
		
		display.getRenderer().setBackgroundColor(ColorRGBA.white);
		
		walkable = new short[sizeX][sizeY];
		
		BufferedReader br = null;

		try {
			br = new BufferedReader(new InputStreamReader(new FileInputStream("media/maps/inpass.txt")));
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			System.exit(1);
		}
		
		int value = -1;
		String line = null;
		
		for (int y = 0; y < sizeY; y++) {
			try {
				line = br.readLine();
			} catch (IOException e) {
				e.printStackTrace();
				System.exit(1);
			}
			for (int x = 0; x < sizeX; x++) {
				value = Integer.parseInt(line.substring(x,x+1));
				if (value == 0) {
					walkable[x][sizeY - (y + 1)] = 0;
					Quad quad = new Quad("q"+x+"x"+(sizeY - (y + 1)), 10, 10);
					quad.setSolidColor(ColorRGBA.black);
					quad.setLocalTranslation(new Vector3f(x*10, (sizeY - (y + 1))*10, 0));
					rootNode.attachChild(quad);
				} else if (value == 1) {
					walkable[x][sizeY - (y + 1)] = 1;
					Quad quad = new Quad("q"+x+"x"+(sizeY - (y + 1)), 10, 10);
					quad.setSolidColor(ColorRGBA.blue);
					quad.setLocalTranslation(new Vector3f(x*10, (sizeY - (y + 1))*10, 0));
					rootNode.attachChild(quad);
				}
			}
		}
		
		for (int y = 0; y <= sizeY; y++) {
			Line l = new Line("lv"+y, new Vector3f[]{new Vector3f(-5,y*10-5,0.1f), new Vector3f(sizeX*10-5,y*10-5,0.1f)}, null, null, null);
			l.setSolidColor(ColorRGBA.gray);
			rootNode.attachChild(l);
		}

		for (int x = 0; x <= sizeX; x++) {
			Line l = new Line("lh"+x, new Vector3f[]{new Vector3f(x*10-5,-5,0.1f), new Vector3f(x*10-5,sizeY*10-5,0.1f)}, null, null, null);
			l.setSolidColor(ColorRGBA.gray);
			rootNode.attachChild(l);
		}
		
		Sphere startPoint = new Sphere("startPoint", new Vector3f(start.x*10,start.y*10,0), 10, 10, 2);
		startPoint.setSolidColor(ColorRGBA.green);
		rootNode.attachChild(startPoint);

		Sphere destPoint = new Sphere("destPoint", new Vector3f(dest.x*10,dest.y*10,0), 10, 10, 2);
		destPoint.setSolidColor(ColorRGBA.red);
		rootNode.attachChild(destPoint);
		
		calculateBasicAStar();

		lightState.setEnabled(false);
		
		cam.setLocation(new Vector3f(50,50,250));

		// IMPORTANT
		rootNode.updateRenderState();
		
//System.exit(0);
	}

	@Override
	protected void simpleRender() {
		super.simpleRender();
	}

	@Override
	protected void simpleUpdate() {
		super.simpleUpdate();
	}

	private void calculateBasicAStar() {
		long debut = new Date().getTime();
		listeOuverte = new Hashtable<String, PathTile>();
		listeFermee = new Hashtable<String, PathTile>();
		
		PathTile startTile = new PathTile(null, (int)start.x, (int)start.y, 0, calculateManhattanDistance((int)start.x, (int)start.y, (int)dest.x, (int)dest.y));
		listeOuverte.put(startTile.getName(), startTile);
		treatOpenList();
		if (tileDest == null) {
			LOGGER.error("Impossible d'atteindre la destination !");
		} else {
			// TODO : remonter la chaine !
			LOGGER.debug(tileDest);
			Sphere s = new Sphere("s"+tileDest.getName(), new Vector3f(tileDest.getX()*10,tileDest.getY()*10,0), 10, 10, 2);
			s.setSolidColor(ColorRGBA.lightGray);
			rootNode.attachChild(s);
			while (tileDest.getParent() != null) {
				LOGGER.debug(tileDest.getParent());
				s = new Sphere("s"+tileDest.getName(), new Vector3f(tileDest.getX()*10,tileDest.getY()*10,0), 10, 10, 2);
				s.setSolidColor(ColorRGBA.lightGray);
				rootNode.attachChild(s);
				tileDest = tileDest.getParent();
			}
		}
		LOGGER.info((new Date().getTime() - debut)+" ms");
	}
	
	private void treatOpenList() {
		int count = 0;
		
		while (! listeOuverte.isEmpty() && count < 300) {
			int minF = Integer.MAX_VALUE;
			PathTile curTile = null;
			
			LOGGER.debug("\nliste ouverte : "+listeOuverte);

			for (Iterator<Entry<String,PathTile>> it = listeOuverte.entrySet().iterator(); it.hasNext();) {
				PathTile tile = it.next().getValue();
				if (tile.getF() < minF) {
					minF = tile.getF();
					curTile = tile;
				}
			}
			
			LOGGER.debug("curTile : "+curTile);

			listeOuverte.remove(curTile.getName());
			listeFermee.put(curTile.getName(), curTile);
			
			if (curTile.getX() == dest.x && curTile.getY() == dest.y) {
				tileDest = curTile;
				LOGGER.info("Destination atteinte !\n");
				return;
			}
			
			PathTile[] voisins = findNeighbours(curTile, (int)dest.x, (int)dest.y);
			
			int newG = 0;
			PathTile oldTile = null;
	
			for (int i = 0; i < voisins.length; i++) {
				if (! listeFermee.containsKey(voisins[i].getName())) {
					LOGGER.debug("voisin : "+voisins[i]);
					if (listeOuverte.containsKey(voisins[i].getName())) {
						newG = curTile.getG();
						if (voisins[i].getX() == curTile.getX() + 1 ||
								voisins[i].getX() == curTile.getX() - 1 ||
								voisins[i].getY() == curTile.getY() + 1 ||
								voisins[i].getY() == curTile.getY() - 1) {
							newG += DIAG_COST;
						} else {
							newG += STRAIGHT_COST;
						}
						oldTile = listeOuverte.get(voisins[i].getName());
						if (newG < oldTile.getG()) {
							oldTile.setParent(curTile);
							oldTile.setG(newG);
							listeOuverte.put(oldTile.getName(), oldTile);
							LOGGER.debug("old : "+oldTile);
						}
					} else {
						listeOuverte.put(voisins[i].getName(), voisins[i]);
						LOGGER.debug("add : "+voisins[i]);
					}
				} else {
					LOGGER.debug("voisin ferm : "+voisins[i]);
				}
			}
			
			count++;
		}
	}
	
	private PathTile[] findNeighbours(PathTile curTile, int xDest, int yDest) {
		Vector<PathTile> tiles = new Vector<PathTile>();
		int x = curTile.getX();
		int y = curTile.getY();
		int g = curTile.getG();
		
		if (x == 0 && y == 0) {
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
			if (walkable[x+1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y+1,  g + DIAG_COST, calculateManhattanDistance(x+1, y+1, xDest, yDest)));
			}
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
		} else if (x == 0 && y == sizeY -1) {
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
			if (walkable[x+1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y-1,  g + DIAG_COST, calculateManhattanDistance(x+1, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
		} else if (x == sizeX -1 && y == 0) {
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y+1,  g + DIAG_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
		} else if (x == sizeX -1 && y == sizeY -1) {
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y-1,  g + DIAG_COST, calculateManhattanDistance(x-1, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
		} else if (x == 0) {
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y+1,  g + DIAG_COST, calculateManhattanDistance(x+1, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
			if (walkable[x+1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y-1,  g + DIAG_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
		} else if (x == sizeX -1) {
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
			if (walkable[x-1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y+1,  g + DIAG_COST, calculateManhattanDistance(x-1, y+1, xDest, yDest)));
			}
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y-1,  g + DIAG_COST, calculateManhattanDistance(x-1, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
		} else if (y == 0) {
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y+1,  g + DIAG_COST, calculateManhattanDistance(x-1, y+1, xDest, yDest)));
			}
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y+1,  g + DIAG_COST, calculateManhattanDistance(x+1, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
		} else if (y == sizeY -1) {
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y-1,  g + DIAG_COST, calculateManhattanDistance(x-1, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
			if (walkable[x+1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y-1,  g + DIAG_COST, calculateManhattanDistance(x+1, y-1, xDest, yDest)));
			}
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
		} else {
			if (walkable[x-1][y] == 0) {
				tiles.add(new PathTile(curTile, x-1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x-1, y, xDest, yDest)));
			}
			if (walkable[x-1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y+1,  g + DIAG_COST, calculateManhattanDistance(x-1, y+1, xDest, yDest)));
			}
			if (walkable[x][y+1] == 0) {
				tiles.add(new PathTile(curTile, x, y+1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y+1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y+1,  g + DIAG_COST, calculateManhattanDistance(x+1, y+1, xDest, yDest)));
			}
			if (walkable[x+1][y] == 0) {
				tiles.add(new PathTile(curTile, x+1, y,  g + STRAIGHT_COST, calculateManhattanDistance(x+1, y, xDest, yDest)));
			}
			if (walkable[x+1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x+1, y-1,  g + DIAG_COST, calculateManhattanDistance(x+1, y-1, xDest, yDest)));
			}
			if (walkable[x][y-1] == 0) {
				tiles.add(new PathTile(curTile, x, y-1,  g + STRAIGHT_COST, calculateManhattanDistance(x, y-1, xDest, yDest)));
			}
			if (walkable[x-1][y-1] == 0) {
				tiles.add(new PathTile(curTile, x-1, y-1,  g + DIAG_COST, calculateManhattanDistance(x-1, y-1, xDest, yDest)));
			}
		}
		
		PathTile[] result = new PathTile[tiles.size()];
		result = tiles.toArray(result);
		return (result);
	}
	
	private int calculateManhattanDistance(int xOri, int yOri, int xDest, int yDest) {
		return (((xDest>xOri?xDest-xOri:xOri-xDest) + (yDest>yOri?yDest-yOri:yOri-yDest)) * STRAIGHT_COST);
	}
	
}

