package fr.bhprod.sga.terrain.layer;

import com.jme.bounding.BoundingBox;
import com.jme.image.Texture;
import com.jme.math.Vector2f;
import com.jme.math.Vector3f;
import com.jme.renderer.ColorRGBA;
import com.jme.renderer.Renderer;
import com.jme.scene.TriMesh;
import com.jme.scene.state.AlphaState;
import com.jme.scene.state.TextureState;
import com.jme.util.geom.BufferUtils;

import fr.bhprod.sga.terrain.Terrain;

/**
 * 
 * @author jbeaufils
 * @version 1.0
 */

public final class Patch {

	public static final TriMesh createPatch(String name, int patchSize, int tx,
			int tz, Texture texture, boolean textureWrap, boolean alpha, boolean inpass,
			Terrain terrain, Renderer renderer) {

		int patchSize1 = patchSize + 1;
		int size_square = patchSize1 * patchSize1;

		Vector3f[] vertexes = new Vector3f[size_square];
		Vector3f[] normals = new Vector3f[size_square];
		Vector2f[] texCoords = new Vector2f[size_square];
		ColorRGBA[] colors = null;
		if (inpass) {
			colors = new ColorRGBA[size_square];
		}

		// 2 tris for each patch * 3 vertices (6)
		int[] indexes = new int[patchSize * patchSize * 6];

		// For each tile
		for (int z = 0; z < patchSize; z++) {
			for (int x = 0; x < patchSize; x++) {
				int i = z * patchSize * 6 + (x * 6);
				indexes[i] = z * patchSize1 + x;
				indexes[i + 1] = z * patchSize1 + x + 1;
				indexes[i + 2] = (z + 1) * patchSize1 + x;
				indexes[i + 3] = z * patchSize1 + x + 1;
				indexes[i + 4] = (z + 1) * patchSize1 + x + 1;
				indexes[i + 5] = (z + 1) * patchSize1 + x;
			}
		}

		// For each vertex
		for (int z = 0; z < patchSize1; z++) {
			for (int x = 0; x < patchSize1; x++) {
				vertexes[z * patchSize1 + x] = new Vector3f((x + tx)
						* terrain.getTileRealSize(), terrain.getHeights()[x + tx][z + tz], -(z + tz)
						* terrain.getTileRealSize());
				normals[z * patchSize1 + x] = new Vector3f(0, 0, 1);
				if (textureWrap) {
					texCoords[z * patchSize1 + x] = new Vector2f(
							(float) (tx + x) / terrain.getVerticesX(), (float) (tz + z)
									/ terrain.getVerticesZ());
				} else {
					texCoords[z * patchSize1 + x] = new Vector2f(
							(float) x * 0.2f, (float) z * 0.2f);
				}
				if (inpass) {
					if ((x+tx) < terrain.getTilesX() && (z + tz) < terrain.getTilesZ() && terrain.getInpassMap()[x + tx][z + tz] != 0) {
						colors[z * patchSize1 + x] = new ColorRGBA(1,0,0,0.9f);
					} else {
						colors[z * patchSize1 + x] = new ColorRGBA(1,1,1,1);
					}
				}
			}
		}

		TriMesh mesh = new TriMesh(name + "[" + (tx / patchSize) + "]["
				+ (tz / patchSize) + "]");

		TextureState ts = renderer.createTextureState();
		ts.setTexture(texture);
		ts.setEnabled(true);
		mesh.setRenderState(ts);

		if (alpha) {
			AlphaState as = renderer.createAlphaState();
			as.setBlendEnabled(true);
			as.setDstFunction(AlphaState.SB_ONE_MINUS_SRC_ALPHA);
			as.setSrcFunction(AlphaState.DB_ONE);
			as.setTestEnabled(true);
			as.setTestFunction(AlphaState.TF_GREATER);
			as.setEnabled(true);
			mesh.setRenderState(as);
		}

		if (inpass) {
			mesh.reconstruct(BufferUtils.createFloatBuffer(vertexes), BufferUtils
					.createFloatBuffer(normals), BufferUtils
					.createFloatBuffer(colors), BufferUtils
					.createFloatBuffer(texCoords), BufferUtils
					.createIntBuffer(indexes));
		} else {
			mesh.reconstruct(BufferUtils.createFloatBuffer(vertexes), BufferUtils
			.createFloatBuffer(normals), null, BufferUtils
			.createFloatBuffer(texCoords), BufferUtils
			.createIntBuffer(indexes));
		}
		
		mesh.setModelBound(new BoundingBox());
		mesh.updateModelBound();

		return (mesh);
	}

}
